#!/usr/bin/env python
# @copyright (c) 2002-2016 Acronis International GmbH. All rights reserved.
# EULA: https://www.acronis.com/en-us/download/docs/eula/corporate/
import json
import platform
import shutil
import tempfile
from contextlib import contextmanager
from typing import Tuple
from uuid import uuid4

import is_lsr
import logging
import os
import subprocess
from baas_lsr_support import find_machine_backups, CommonBoxName
from colorama import init as colorama_init
from colorlog.escape_codes import parse_colors
from is_lsr import configuration
from lxml import etree


log = logging.getLogger(__name__)


class NoOrderError(is_lsr.Error):
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return 'No Large Scale Recovery orders are ready for processing: ' + self.text

    @staticmethod
    def create(registrations, box):
        order = [registration for registration in registrations if lsr_order(registration, box)]
        if order:
            return NoOrderError('the subaccount {0} order is in state {1}'.format(box, order[0]['StatusId']))
        else:
            return NoOrderError('no orders available for the subaccount ' + box)


@contextmanager
def certificate_file(certificate_data: bytes) -> tempfile.NamedTemporaryFile:
    filename = '{}.pem'.format(str(uuid4()))
    with open(filename, 'wb+') as certificate:
        certificate.write(certificate_data)
        certificate.flush()
        certificate.seek(0)
        certificate.close()

    filename = os.path.abspath(filename)

    try:
        yield filename
    finally:
        os.remove(filename)


def lsr_order(registration, box):
    return registration['ProductType'] == 'LSR' and registration['SubaccountId'] == box


def suitable_for_lsr(registration, box):
    return lsr_order(registration, box) and registration['StatusId'] in ['103', '106']


def mark_as_lsr(file):
    tree = etree.parse(file, parser=etree.XMLParser(recover=True))
    root = tree.getroot()
    is_lsr = root.find('.//metainfo[@type="archive"]/is-lsr')
    if is_lsr is not None:
        is_lsr.text = is_lsr.text.replace('0', '1')
    else:
        archive_metainfo = root.find('.//metainfo[@type="archive"]')
        if archive_metainfo is not None:
            is_lsr = etree.SubElement(archive_metainfo, 'is-lsr')
            is_lsr.text = '1'
    return tree


def get_around_lsr_meta_reindexing(path):
    meta_files = [f for f in os.listdir(path) if os.path.splitext(f)[1] == '.xml']
    for filename in meta_files:
        meta_path = os.path.join(path, filename)
        try:
            with open(meta_path, encoding='utf-8') as file:
                updated_xml = mark_as_lsr(file)

            if updated_xml:
                updated_xml.write(meta_path)
        except (OSError, etree.XMLSyntaxError):
            logging.debug('Failed to modify metafile:', exc_info=True)


class LSR(object):
    def __init__(self, rpc_uri, account, box):
        self.rpc_uri = rpc_uri
        self.account = account
        self.box = box

    def perform(self, path):
        server, port, cert = self.get_access_to_server()
        logging.info('Using %s:%s for download', server, port)

        self.change_status(104, "Changing LSR status to 'Writing data'... ")

        self.print_starting_information()

        self.run_is_tool(server, port, cert, path)

        get_around_lsr_meta_reindexing(path)

        self.change_status(105, "Changing LSR status to 'Writing data has been completed'... ")
        logging.info(parse_colors('bold_green') + 'The data has been written to the media. Do not forget to deliver the media to the customer.' + parse_colors('reset'))

    def run_is_tool(self, server, port, cert, path):
        try:
            lsr_parameters = ['--lsr', '--server', server, '--port', str(port), '--brand', '1', '--account',
                              self.account, '--box', self.box] + is_lsr.deduce_vault_parameter(path)
            is_lsr.run_is_tool(lsr_parameters, cert)
        except:
            self.change_status(106, "Changing LSR status to 'The order is on hold'... ")
            raise


class AbrLSR(LSR):
    def __init__(self, rpc_uri, hash, insecure):
        _, self.email, self.password_hash, self.box, account = hash.split()
        super().__init__(rpc_uri, account, self.box)
        self.insecure = insecure
        registrations = is_lsr.request_registrations(self.email, self.password_hash, rpc_uri, insecure=insecure)
        lsr_registration = [registration for registration in registrations if suitable_for_lsr(registration, self.box)]
        if not lsr_registration:
            raise NoOrderError.create(registrations, self.box)
        self.order_id = lsr_registration[0]['id']

    def get_access_to_server(self):
        return is_lsr.get_access_to_abr_server(self.rpc_uri, self.email, self.password_hash, self.insecure)

    def change_status(self, status, message):
        return is_lsr.change_status(self.rpc_uri, self.email, self.box, self.password_hash, self.order_id, status, message, insecure=self.insecure)

    def print_starting_information(self):
        logging.info('Writing data for subaccount %s of account %s (%s) by order %s', self.box, self.account, self.email, self.order_id)


class BaaSLSR:
    def __init__(self, hash: str, username: str, password: str, insecure: bool):
        _, account, machine_id, rpc = hash.split()
        account = account.partition(':')[2]
        self.machine_id = machine_id.partition(':')[2]
        rpc = rpc.partition(':')[2]
        self.rpc_uri = is_lsr.get_baas_rpc({'rain': rpc})
        self.account = account
        self.subaccount = None
        try:
            self.subaccount = str(
                is_lsr.get_machine_subaccount(self.rpc_uri, account, self.machine_id, username, password, insecure)
            )
        except:
            log.warning('Failed to retrieve subaccount for given machine. Will work with default: %s', CommonBoxName)

        self.server, self.port, self.cert = is_lsr.get_access_to_baas_server(
            self.rpc_uri, self.account, **{'email': username, 'password': password, 'insecure': insecure}
        )

    def run_archive_tool(self, tool_name: str, *opts) -> Tuple[str, str]:
        tool_path = os.path.join(os.path.dirname(is_lsr.is_tool_path()), tool_name)
        if platform.system().lower().startswith('linux'):
            tool_path = shutil.which(tool_name)
            if tool_path is None:
                raise Exception('archive3.rpm not installed - binaries not found')

        with certificate_file(self.cert) as cert_file:
            command = [
                tool_path,
                '--cert', cert_file,
                '--astor', str(self.server),
            ] + list(opts)

            log.info('Executing util: %s', ' '.join(command))

            process = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                close_fds=False,
                env=os.environ
            )

            results = process.communicate()

        stdout, stderr = tuple(map(lambda x: x.decode('utf-8'), results))

        log.debug('Util %s logs: %s', tool_name, stderr)
        return stdout, stderr

    def get_archive3_backups_list(self) -> list:
        log.info('Getting backups list')
        stdout, stderr = self.run_archive_tool(
            'archive_ctl',
            '--list-archives', '/{}'.format(CommonBoxName)
        )
        result = stderr.split('\n')[1:-2]
        backups = list(map(lambda x: x.split(maxsplit=2)[2], result))
        backups = [backup.strip() for backup in backups if 'tibx' in backup]

        given_machine_backups = []
        for backup in backups:
            try:
                stdout, stderr = self.run_archive_tool(
                    'archive_ctl',
                    '--file', '/{}/{}'.format(CommonBoxName, backup),
                    '--info'
                )
                meta = json.loads(stdout)['meta']
                if meta['agent_machine_id'].lower() == self.machine_id.lower():
                    given_machine_backups.append(backup)
            except KeyError:
                log.error('Skipping backup %s: meta is broken. Meta: %r', backup, json.loads(stdout))

        log.info('Backups for given machine: %s', given_machine_backups)
        return given_machine_backups

    def download_archive3_backup(self, backup_name: str, destination: str) -> None:
        log.info('Downloading backup %s to %s', backup_name, destination)
        uploader_opts = [
            'archive_io_ctl',
            '--copy', '/{}/{}'.format(CommonBoxName, backup_name),
            '--dst', os.path.join(destination, backup_name),
            '--show-progress',
            '--infinite-timeouts'
        ]
        if os.path.exists(os.path.join(destination, backup_name)):
            uploader_opts.append('--continue')

        self.run_archive_tool(*uploader_opts)

        log.info('Backup %s successfully downloaded', backup_name)

    def perform(self, path: str):
        log.info('Processing backups...')
        backups = self.get_archive3_backups_list()
        for backup in backups:
            try:
                self.download_archive3_backup(backup, path)
            except:
                log.exception('Failed to download backup %s')

        log.info('Processing legacy backups...')
        self.process_archive2_backups(self.server, self.port, self.cert, path)

    def process_archive2_backups(self, server, port, cert, path):
        log.info('Downloading legacy data...')
        if self.subaccount is not None:
            self.download_from_subaccount(server, port, cert, path)

        backups = find_machine_backups(server, int(port), cert, self.machine_id)
        backups = [backup for backup in backups if backup]
        log.info('Found %s legacy backup(s) related to machine %s', len(backups), self.machine_id)
        [self.download_archive2_backups(backup, server, port, cert, path) for backup in backups]
        get_around_lsr_meta_reindexing(path)

    def download_archive2_backups(self, backup, server, port, cert, path):
        try:
            lsr_parameters = ['--lsr', '--server', server, '--port', str(port), '--brand', '1', '--account',
                              self.account, '--box', CommonBoxName] + is_lsr.deduce_vault_parameter(path)
            for file in backup:
                lsr_parameters.append('--file')
                lsr_parameters.append(file)
            is_lsr.run_is_tool(lsr_parameters, cert)
        except:
            log.exception('Failed to download legacy backup %s', backup)

    def download_from_subaccount(self, server, port, cert, path):
        log.info('Downloading legacy backups from subaccount %s', self.subaccount)
        try:
            lsr_parameters = ['--lsr', '--server', server, '--port', str(port), '--brand', '1', '--account',
                              self.account, '--box', self.subaccount] + is_lsr.deduce_vault_parameter(path)
            is_lsr.run_is_tool(lsr_parameters, cert)
        except:
            log.exception('Error occurred while running is_tool')
            raise

        log.info('Downloading legacy backups from subaccount %s finished', self.subaccount)


def main(args):
    import sys
    if args.hash is None:
        logging.error("The hash is invalid.")
        sys.exit(1)

    try:
        hash = args.hash.decode('utf-8')

        if hash.startswith('LSR_request'):
            if args.password:
                password = args.password
            else:
                from getpass import getpass
                password = getpass()
            processor = BaaSLSR(hash, args.username, password, args.insecure)
        else:
            processor = AbrLSR(is_lsr.get_abr_rpc(configuration()), hash, args.insecure)

        if not os.path.isdir(args.path):
            raise Exception('Destination is not a directory')

        processor.perform(args.path)
    except KeyboardInterrupt:
        logging.error("\nOperation was cancelled.\n")
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        logging.error("An error has occurred.")
        sys.exit(e.returncode)
    except Exception as e:
        logging.error("An error has occurred: " + str(e))
        logging.debug("Exception: ", exc_info=True)
        sys.exit(1)


def parse_arguments():
    import argparse
    parser = argparse.ArgumentParser(add_help=True, description='Tool to download Large Scale Recovery backups from Acronis storage.')
    parser.add_argument('--username', required=False, help="service account that performs LSR")
    parser.add_argument('--password', required=False, help="service account credentials to perform LSR")
    parser.add_argument('--insecure', dest='insecure', action='store_true', help='allows connection to a server with a not trusted SSL certificate')
    parser.add_argument('hash', metavar='<hash>', type=is_lsr.Base64Type(), help='a hash code that must be obtained from the Large Scale Recovery management panel')
    parser.add_argument('path', metavar='<path>', help='path where lsr backups will be copied')
    parser.set_defaults(insecure=False)
    return parser.parse_args()


if __name__ == "__main__":
    colorama_init()
    from pretty_logging import setup_logging
    setup_logging()
    main(parse_arguments())
